@@ -15,6 +15,9 @@ use aliased 'LWP::UserAgent' => '_UserAgent';
 use Carp qw/
     croak
 /;
+use URI::Escape qw/
+    uri_escape
+/;
 use XML::Simple qw/
     XMLin
     XMLout
@@ -24,24 +27,23 @@ use XML::Simple qw/
 
 WWW::PivotalTracker - Functional interface to Pivotal Tracker L<http://www.pivotaltracker.com/>
 
-=head1 VERSION
-
-Version 0.1.0_5
-
 =cut
 
-use version; our $VERSION = qv("0.1.0_5");
+our $VERSION = "0.16_01";
 
 =head1 SYNOPSIS
 
 This module provides simple methods to interact with the Pivotal Tracker API.
 
     use WWW::PivotalTracker qw/
+        add_note
         add_story
         all_stories
         delete_story
         project_details
         show_story
+        stories_for_filter
+        update_story
     /;
 
     my $details = project_details("API Token", "Project ID");
@@ -55,11 +57,14 @@ Nothing is exported by default.  See B<FUNCTIONS> for the exportable functions.
 =cut
 
 our @EXPORT_OK = qw/
+    add_note
     add_story
     all_stories
     delete_story
     project_details
     show_story
+    stories_for_filter
+    update_story
 /;
 
 #This should never contain anything.  It's just here to make sure that :all
@@ -86,25 +91,20 @@ iteration, and which day of the week the iterations start on.
 
 =cut
 
-sub project_details($token, $project_id) {
+sub project_details($token, $project_id)
+{
     croak("Malformed Project ID: '$project_id'") unless __PACKAGE__->_check_project_id($project_id);
 
     my $response = __PACKAGE__->_do_request($token, "projects/$project_id", "GET");
 
-
-    if (!defined $response || lc $response->{'success'} ne 'true') {
-        return {
-            success => 'false',
-            errors  => (defined $response && exists $response->{'errors'} ? $response->{'errors'} : 'Epic fail!'),
-        };
-    }
+    return { success => 'false' } if (!defined $response);
 
     return {
         success         => 'true',
-        iteration_weeks => $response->{'project'}->{'iteration_length'}->{'content'},
-        name            => $response->{'project'}->{'name'},
-        point_scale     => $response->{'project'}->{'point_scale'},
-        start_day       => $response->{'project'}->{'week_start_day'},
+        iteration_weeks => $response->{'iteration_length'}->{'content'},
+        name            => $response->{'name'},
+        point_scale     => $response->{'point_scale'},
+        start_day       => $response->{'week_start_day'},
     };
 }
 
@@ -142,17 +142,12 @@ sub show_story($token, $project_id, $story_id)
 
     my $response = __PACKAGE__->_do_request($token, "projects/$project_id/stories/$story_id", "GET");
 
-    if (!defined $response || lc $response->{'success'} ne 'true') {
-        return {
-            success => 'false',
-            errors  => (defined $response && exists $response->{'errors'} ? $response->{'errors'} : 'Epic fail!'),
-        };
-    }
+    return { success => 'false' } if (!defined $response);
 
-    my $story = __PACKAGE__->_sanitize_story_xml($response->{'story'}->[0]);
+    my $story = __PACKAGE__->_sanitize_story_xml($response);
 
     return {
-        success       => 'true',
+        success => 'true',
         %{$story},
     };
 }
@@ -169,14 +164,9 @@ sub all_stories($token, $project_id)
 
     my $response = __PACKAGE__->_do_request($token, "projects/$project_id/stories", "GET");
 
-    if (!defined $response || lc $response->{'success'} ne 'true') {
-        return {
-            success => 'false',
-            errors  => (defined $response && exists $response->{'errors'} ? $response->{'errors'} : 'Epic fail!'),
-        };
-    }
+    return { success => 'false' } if (!defined $response);
 
-    my @stories = map { __PACKAGE__->_sanitize_story_xml($_) } @{$response->{'stories'}->{'story'}};
+    my @stories = map { __PACKAGE__->_sanitize_story_xml($_) } @{$response->{'story'}};
 
     return {
         success => 'true',
@@ -249,14 +239,9 @@ sub add_story($token, $project_id, $story_details)
 
     my $response = __PACKAGE__->_do_request($token, "projects/$project_id/stories", "POST", $content);
 
-    if (!defined $response || $response->{'success'} ne 'true') {
-        return {
-            success => 'false',
-            errors  => $response->{'errors'},
-        };
-    }
+    return { success => 'false' } if (!defined $response);
 
-    my $story = __PACKAGE__->_sanitize_story_xml($response->{'story'}->[0]);
+    my $story = __PACKAGE__->_sanitize_story_xml($response);
 
     return {
         success       => 'true',
@@ -271,7 +256,7 @@ Delete an existing story.
     my $result = delete_story($token, $project_id, $story_id);
 
     print $result->{'success'};
-    print $result->{'message'};
+    print $result->{'name'};
 
 =cut
 
@@ -282,17 +267,132 @@ sub delete_story($token, $project_id, $story_id)
 
     my $response = __PACKAGE__->_do_request($token, "projects/$project_id/stories/$story_id", "DELETE");
 
-    if (!defined $response || $response->{'success'} ne 'true') {
-        return {
-            success => 'false',
-            errors  => $response->{'errors'},
-        };
-    }
+    return { success => 'false' } if (!defined $response);
 
-    my $message = $response->{'message'};
     return {
         success => 'true',
-        message => $message,
+        name => $response->{'name'},
+    };
+}
+
+=head2 stories_for_filter
+
+Find all stories given search parameters.
+
+    my $result = stories_for_filter($token, $project_id, $search_filter);
+
+    my @stories;
+    if($result->{'success'} eq 'true') {
+        print $result->{'message'} . "\n";
+
+        @stories = @{$result->{'stories'}};
+    }
+
+In the example above C<< @stories >> will be an array of story hashrefs.  See the description of C<show_story> for the details of the hashrefs.
+
+Any multi-word terms in the search filter must be enclosed by double quotes. (See L<http://www.pivotaltracker.com/help>: Search)
+
+    Example:
+        requester:"Jacob Helwig"
+        owner:"Jacob Helwig"
+        mywork:"Jacob Helwig"
+        state:unstarted
+        type:Feature
+
+=cut
+
+sub stories_for_filter($token, $project_id, $search_filter)
+{
+    croak("Malformed Project ID: '$project_id'") unless __PACKAGE__->_check_project_id($project_id);
+
+    my $response = __PACKAGE__->_do_request($token, "projects/$project_id/stories?filter=" . uri_escape($search_filter), "GET");
+
+    return { success => 'false' } if (!defined $response);
+
+    my @stories = map { __PACKAGE__->_sanitize_story_xml($_) } @{$response->{'story'}};
+
+    return {
+        success => 'true',
+        filter  => $response->{'filter'},
+        stories => [ @stories ],
+    };
+}
+
+=head2 update_story
+
+Update aspects of a given story.
+
+    my $result = update_story($token, $project_id, $story_id, { current_state => 'started' });
+
+See the description of C<show_story> for the details of C<$result>.
+
+=cut
+
+sub update_story($token, $project_id, $story_id, $story_details)
+{
+    croak("Malformed Project ID: '$project_id'") unless __PACKAGE__->_check_project_id($project_id);
+    croak("Malformed Story ID: '$story_id'") unless __PACKAGE__->_check_story_id($story_id);
+
+    foreach my $key (keys %$story_details) {
+        croak("Unrecognized option: $key")
+            unless __PACKAGE__->_is_one_of($key, [qw/
+                created_at
+                current_state
+                deadline
+                description
+                estimate
+                labels
+                name
+                note
+                owned_by
+                requested_by
+                story_type
+            /]);
+    }
+
+    my $content = __PACKAGE__->_make_xml({ story => $story_details });
+
+    my $response = __PACKAGE__->_do_request($token, "projects/$project_id/stories/$story_id", "PUT", $content);
+
+    return { success => 'false' } if (!defined $response);
+
+    my $story = __PACKAGE__->_sanitize_story_xml($response);
+
+    return {
+        success => 'true',
+        %{$story},
+    };
+}
+
+=head2 add_note
+
+Add a note to an existing story.
+
+    my $result = add_comment($token, $project_id, $story_id, $note);
+
+See the description of C<show_story> for the details of C<$result>.
+
+=cut
+
+sub add_note($token, $project_id, $story_id, $note)
+{
+    croak("Malformed Project ID: '$project_id'") unless __PACKAGE__->_check_project_id($project_id);
+    croak("Malformed Story ID: '$story_id'") unless __PACKAGE__->_check_story_id($story_id);
+
+    my $response = __PACKAGE__->_do_request(
+        $token,
+        "projects/$project_id/stories/$story_id/notes",
+        "POST",
+        __PACKAGE__->_make_xml({ note => { text => $note } })
+    );
+
+    return { success => 'false' } if (!defined $response);
+
+    my $new_note = __PACKAGE__->_sanitize_note_xml($response);
+
+    return {
+        success => 'true',
+        %{$new_note},
     };
 }
 
@@ -316,14 +416,10 @@ sub _sanitize_story_xml($class, $story)
     my $labels = undef;
     my $notes = undef;
 
-    $labels = $story->{'labels'}->{'label'} if exists $story->{'labels'};
+    $labels = [ split(',', $story->{'labels'}) ] if exists $story->{'labels'};
     $notes = [
-        map +{
-            id     => $_->{'id'}->{'content'},
-            author => $_->{'author'},
-            date   => $_->{'date'},
-            text   => $_->{'text'},
-        }, @{$story->{'notes'}->{'note'}}
+        map { $class->_sanitize_note_xml($_) }
+            @{$story->{'notes'}->{'note'}}
     ] if exists $story->{'notes'};
 
     return {
@@ -332,19 +428,32 @@ sub _sanitize_story_xml($class, $story)
         description   => $story->{'description'},
         estimate      => $story->{'estimate'}->{'content'},
         current_state => $story->{'current_state'},
-        created_at    => $story->{'created_at'},
-        deadline      => $story->{'deadline'},
+        created_at    => $story->{'created_at'}->{'content'},
+        accepted_at   => $story->{'accepted_at'}->{'content'},
+        deadline      => $story->{'deadline'}->{'content'},
         story_type    => $story->{'story_type'},
         requested_by  => $story->{'requested_by'},
+        owned_by      => $story->{'owned_by'},
         labels        => $labels,
         notes         => $notes,
         url           => $story->{'url'},
     };
 }
 
+sub _sanitize_note_xml($class, $note)
+{
+    return {
+        id     => $note->{'id'}->{'content'},
+        author => $note->{'author'},
+        date   => $note->{'noted_at'}->{'content'},
+        text   => $note->{'text'},
+    };
+    return $note;
+}
+
 sub _do_request($class, $token, $request_url, $request_method; $content)
 {
-    my $base_url = "https://www.pivotaltracker.com/services/v1/";
+    my $base_url = "https://www.pivotaltracker.com/services/v3/";
 
     my $request = _Request->new(
         $request_method,
@@ -356,7 +465,8 @@ sub _do_request($class, $token, $request_url, $request_method; $content)
         $content
     );
 
-    my $response = $class->_post_request($request);
+    my $response = eval { $class->_post_request($request) };
+    return if $@;
 
     return XMLin(
         $response,
@@ -367,10 +477,6 @@ sub _do_request($class, $token, $request_url, $request_method; $content)
             note
             story
         /],
-        GroupTags => {
-            errors => 'error',
-            labels => 'label',
-        },
         KeyAttr => [],
         SuppressEmpty => undef,
     );
@@ -383,7 +489,7 @@ sub _post_request($class, $request)
 
     croak($response->status_line()) unless ($response->is_success());
 
-    return $response;
+    return $response->content();
 }
 
 sub _make_xml($class, HASH $data)
